# -*- coding: utf-8 -*-
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4 import uic
from PyQt4 import QtCore
from PyKDE4.plasma import Plasma
from PyKDE4.kdeui import KIcon,  KConfigDialog,  KDialog
from PyKDE4 import plasmascript
from cpu import *
import string
import dbus
from dbus.mainloop.qt import DBusQtMainLoop


class LabelWithBackground(Plasma.Label):
    
    def __init__(self, parent, args=None):
        Plasma.Label.__init__(self, parent)
        self._showBackground = 0
        
    def setShowBackground(self,  showBackground):
        self._showBackground = showBackground
                
    def paint(self,  painter,  option,  widget=None):
        # paint the background if we want to show it
        if (self._showBackground):
            contentsRect = self.contentsRect()
            painter.save()

            fm = QFontMetrics(painter.font())
            labelWidth = fm.width(' 100%')
            labelHeight = fm.height()

            textRectFactorHor= 1.1
            textRectFactorVer = 1.2
            textRectWidth = int(labelWidth * textRectFactorHor)
            textRectHeight = int(labelHeight * textRectFactorVer)

            textRect = QRect( contentsRect.left() + (contentsRect.width() - textRectWidth) / 2 ,  \
                              contentsRect.top() + (contentsRect.height() - textRectHeight) / 2,  \
                              textRectWidth,  \
                              textRectHeight)

            boxColor = QColor(Qt.white)
            boxColor.setAlpha(48)
            painter.setBrush(boxColor)
            painter.setPen(boxColor)

            roundProp = textRectWidth / textRectHeight
            roundRadius = 40.0
            painter.drawRoundedRect(textRect, float(roundRadius/roundProp), float(roundRadius),  Qt.RelativeSize)

            painter.restore()
        
        # and do the parent painting
        Plasma.Label.paint(self,  painter,  option,  widget)

class CpuAndSystemViewer(plasmascript.Applet):
    
    def __init__(self,parent,args=None):
        plasmascript.Applet.__init__(self,parent)

        self._showChart = 1
        self._showLabelBackground = 1
        self._showSquareAspectRatio = 0
        self._updateInterval = 1000
        
        self._sizeHor = 80
        self._sizeVert = 20
        
        self._firstRun = True



    def init(self):
        # define, has plasmoid configuration interface or not
        self.setHasConfigurationInterface(True)
        self.readConfig()
        
        if self._showSquareAspectRatio:
            self.setAspectRatioMode(Plasma.Square)
        else:
            self.setAspectRatioMode(Plasma.IgnoreAspectRatio)

        # init the widget
        self.initWidget()
        self.initConfigWidgets()
        self.initDataVars()
        self.initToolTip()
                            
        # start the applet
        self.connectToEngine()
        self.update()
                
        
    def tryToFixChartBackground(self):
    # NOTE: make the chart realy adhere to the background. This is a hack.
        if self._showChart:
            # disable chart temporarily and reenable afterwards
            self.setConfigurationInterfaceOptions()
            self.uiChart.chkShowChart.setChecked(False)
            self.configUpdated()            
            self.uiChart.chkShowChart.setChecked(True)
            self.configUpdated()
            
    def initConfigWidgets(self):
        self.widgetGeneral = QWidget()
        self.uiGeneral = uic.loadUi(self.package().filePath('ui', 'configGeneral.ui'), self.widgetGeneral)
        self.widgetChart = QWidget()
        self.uiChart = uic.loadUi(self.package().filePath('ui', 'configChart.ui'), self.widgetChart)


    def initWidget(self):
        self.resize(self._sizeHor,  self._sizeVert)
                
        self.layout = QGraphicsGridLayout(self.applet)
        self.applet.setLayout(self.layout)
        self.setBackgroundHints(Plasma.Applet.DefaultBackground)

        textColor = Plasma.Theme.defaultTheme().color( Plasma.Theme.TextColor ) 
        textColorStr = textColor.name()

        # label
        self.text = LabelWithBackground(self.applet)            
        self.text.setZValue(2)
        self.text.setAlignment(Qt.AlignCenter)
        
        styleSheetStr = """
                                font-weight: bold;
                                color: white;
                        """
        styleSheetStr = styleSheetStr.replace("white",  textColorStr)

        self.text.setStyleSheet(styleSheetStr)        
                                                
        self.text.setShowBackground(self._showLabelBackground)
        self.text.setText('')            
        self.layout.addItem(self.text,  0,  0)

        # chart
        self.chart = self.getChartObject()
        if self._showChart:
            self.layout.addItem(self.chart, 0, 0)
            self.chart.setVisible(True)
        else:
            self.chart.setVisible(False)
            
        self.resize(self._sizeHor,  self._sizeVert)
        # FIXME: showing the chart leads to a square plasmoid for vertical panels even with aspectRatio != Square
        # but only for vertical panels.
            
            
    def getChartObject(self):
        chart = Plasma.SignalPlotter(self.applet)
        chart.setZValue(1)
        chart.setTitle("CPU Usage / Memory Usage")

        chart.setShowTopBar(0)
        chart.setShowLabels(0)
        chart.setShowVerticalLines(0)
        chart.setShowHorizontalLines(0)
        chart.setThinFrame(0)                        # no thin frame lines (bottom + right)
        chart.setSvgBackground(QString(""))          # disable svg background
        chart.setStackPlots(0)                       # NOTE: This makes plot non-filled out
        
        chart.setBackgroundColor(Qt.transparent)     # use transparent background

        chart.setUseAutoRange(0)
        chart.setVerticalRange(0,100)
        
                
        if self._showCpu:   # cpu usage
            chart.addPlot(Qt.green)        

        if self._showMem:  # memory usage
            chart.addPlot(Qt.red)       

        if self._showSwap: # swap usage
            chart.addPlot(Qt.yellow)       
        
        return chart
        
    def setConfigurationInterfaceOptions(self):
        self.uiGeneral.chkShowLabel.setChecked(self._showLabel)
        self.uiGeneral.chkShowLabelBackground.setChecked(self._showLabelBackground)
        self.uiGeneral.spbxUpdateInterval.setValue(self._updateInterval)
        self.uiGeneral.chkIsAspectRatio.setChecked(self._showSquareAspectRatio)
                
        self.uiChart.chkShowChart.setChecked(self._showChart)
        self.uiChart.chkShowCpu.setChecked(self._showCpu)
        self.uiChart.chkShowMem.setChecked(self._showMem)
        self.uiChart.chkShowSwap.setChecked(self._showSwap)
        
        # adapt active / inactive stuff
        self.uiGeneral.chkShowLabelBackground.setEnabled(self._showLabel)
        self.uiChart.chkShowCpu.setEnabled(self._showChart)
        self.uiChart.chkShowMem.setEnabled(self._showChart)
        self.uiChart.chkShowSwap.setEnabled(self._showChart)

            
    def createConfigurationInterface(self,  parent):
        self.initConfigWidgets()

        parent.setButtons(KDialog.ButtonCode(KDialog.Ok | KDialog.Cancel | KDialog.Apply))
        parent.addPage(self.uiGeneral,  "General",  self.icon(),  QString(),  False)
        parent.addPage(self.uiChart,  "Chart",  self.icon(),  QString(),  False)
        
        self.setConfigurationInterfaceOptions()        
        
        self.connect(parent, SIGNAL("applyClicked()"), self.configUpdated)
        self.connect(parent, SIGNAL("okClicked()"), self.configUpdated)
        
    def configUpdated(self):
        if not self.uiGeneral:
            print "Error: No config dialog ui object!"
            return
                                
        # aspect ratio
        ratio = self.uiGeneral.chkIsAspectRatio.isChecked()
        
        if self._showSquareAspectRatio != ratio:
            print "aspect ratio option changed"
            self._showSquareAspectRatio = ratio
                        
            if (self._showSquareAspectRatio):
                self.setAspectRatioMode(Plasma.Square)
            else:
                self.setAspectRatioMode(Plasma.IgnoreAspectRatio)
            self.resize(self._sizeHor,  self._sizeVert)
            
        # label itself
        self._showLabel= self.uiGeneral.chkShowLabel.isChecked()
        if self._showLabel:
            self.text.setVisible(True)
        else:
            self.text.setVisible(False)
                        
        # label background
        self.text.setShowBackground(self._showLabelBackground)
                
        # chart options
        self._showCpu = self.uiChart.chkShowCpu.isChecked()
        self._showMem = self.uiChart.chkShowMem.isChecked()
        self._showSwap = self.uiChart.chkShowSwap.isChecked()

        # handle chart options; 
        # add a new chart object so that any changes to which charts to show are picked up
        showChrt = self.uiChart.chkShowChart.isChecked()
        self.showChart(showChrt)
        self._showChart = showChrt
        
        # interval        
        self._updateInterval = self.uiGeneral.spbxUpdateInterval.value()
        self.connectToEngine()

        self.writeConfig()
        
    def showChart(self,  showChrt):
        if showChrt:
            self.layout.removeAt(1)
            self.chart = self.getChartObject()
            self.layout.addItem(self.chart,  0,  0)
            self.chart.setVisible(True)
        else:
            self.chart.setVisible(False)


    def writeConfig(self):
        cg = self.config()
        
        cg.writeEntry("squareAspectRatio", self._showSquareAspectRatio)
        cg.writeEntry("withLabel", self._showLabel)
        cg.writeEntry("opaqueLabelBackground",  self._showLabelBackground)
        cg.writeEntry("updateInterval",  self._updateInterval)
        cg.writeEntry("withChart",  self._showChart)
        cg.writeEntry("showCpu",  self._showCpu)
        cg.writeEntry("showMem",  self._showMem)
        cg.writeEntry("showSwap",  self._showSwap)
    
        self.emit(SIGNAL("configNeedsSaving()"))
        self.updateConstraints(Plasma.SizeConstraint)
        
                
    def readConfig(self):
        cg = self.config()
        
        self._showSquareAspectRatio = cg.readEntry("squareAspectRatio", False).toBool()
        self._showLabel = cg.readEntry("withLabel",  True).toBool()
        self._showLabelBackground = cg.readEntry("opaqueLabelBackground",  True).toBool()
        self._updateInterval,  ok =  cg.readEntry("updateInterval",  1000).toInt()
        self._showChart = cg.readEntry("withChart",True).toBool()
        self._showCpu = cg.readEntry("showCpu", True).toBool()
        self._showMem = cg.readEntry("showMem", True).toBool()
        self._showSwap = cg.readEntry("showSwap", True).toBool()
        
    def initDataVars(self):
        self._painter = 0
        self._contentsRect = 0
        
        self._dataCpuClock = 0.0
        self._dataCpuIdle = 0.0
        self._dataCpuUsed = 0.0
        self._dataMemUsed = 0.0
        self._dataMemApplications = 0.0
        self._dataMemFree = 0.0
        self._dataSwapUsed = 0.0
        self._dataSwapFree = 0.0

    def initToolTip(self):
        self.tooltipdata = Plasma.ToolTipContent()
        self.tooltipdata.setMainText("System Information")
        self.tooltipdata.setImage(KIcon('view-statistics'))
        Plasma.ToolTipManager.self().registerWidget(self.applet)
        self.updateToolTip()
        
    def connectToEngine(self):
        self.systemEngine = self.dataEngine("systemmonitor")
        self.systemEngine.connectSource("cpu/system/TotalLoad", self, self._updateInterval)
        self.systemEngine.connectSource("cpu/system/AverageClock", self, self._updateInterval)
        self.systemEngine.connectSource("cpu/system/idle", self, self._updateInterval)
        self.systemEngine.connectSource("mem/physical/used", self, self._updateInterval)
        self.systemEngine.connectSource("mem/physical/application", self, self._updateInterval)
        self.systemEngine.connectSource("mem/physical/free", self, self._updateInterval)
        self.systemEngine.connectSource("mem/swap/used", self, self._updateInterval)
        self.systemEngine.connectSource("mem/swap/free", self, self._updateInterval)
        

    @pyqtSignature("dataUpdated(const QString &, const Plasma::DataEngine::Data &)")
    def dataUpdated(self, sourceName, data):

        if not data.has_key(QString("value")):
            # print "No value for source: " + sourceName
            return
            
        if str(sourceName) == "cpu/system/idle":
            self._dataCpuIdle, conversionOk = data[QString("value")].toDouble()
            self._dataCpuUsed = 100.0 - self._dataCpuIdle
            
        elif str(sourceName) == "cpu/system/TotalLoad":
            a = 1

        elif str(sourceName) == "cpu/system/AverageClock":
            self._dataCpuClock, conversionOk = data[QString("value")].toDouble()
            
        elif str(sourceName) == "mem/physical/used":
            self._dataMemUsed, conversionOk = data[QString("value")].toDouble()
            
        elif str(sourceName) == "mem/physical/application":
            self._dataMemApplications, conversionOk = data[QString("value")].toDouble()
            
        elif str(sourceName) == "mem/physical/free":
            self._dataMemFree, conversionOk = data[QString("value")].toDouble()
            
        elif str(sourceName) == "mem/swap/used":
            self._dataSwapUsed, conversionOk = data[QString("value")].toDouble()
            
        elif str(sourceName) == "mem/swap/free":
            self._dataSwapFree, conversionOk = data[QString("value")].toDouble()
            
        else:
            print "No action defined for data from source " + str(sourceName)

        # NOTE: Hack to get background of chart correct
        if self._firstRun:
            print "Fixing chart background. This should happen only once."
            self.tryToFixChartBackground()
            self._firstRun = False

        # update the widget, i. e. let paintInterface be called
        self.update()

    def getLabelText(self):
        cpuUsage = int( self._dataCpuUsed / 5) * 5
        
        # format string nicely
        if cpuUsage < 10:
            t = "  " + str(cpuUsage)

        else:
            t = str(cpuUsage)
            
        text = '' + t + '%'
        return text
        

    def paintInterface(self, painter, option, rect):
        # Store painter and contentsRect
        self._painter = painter
        self._contentsRect = rect
        
        self.updateToolTip()
        # TODO: Call updateToolTip only for shown tooltip. Yet, as no update on tooltipAboutToShow is 
        # done, we would end up with an out-of-date tooltip for the first interval
        # if Plasma.ToolTipManager.self().isVisible(self.applet):
        #     self.updateToolTip()
        
        if self._showChart:
            self.updateChart()

        txt = self.getLabelText()
        self.text.setText(txt)
        
        
    def getMemPercent(self):
        # calculate memory used in percent
        memTotal = (self._dataMemFree + self._dataMemUsed) / 1024
        memUsed = self._dataMemApplications / 1024
        memPercent = 0.0
        if memTotal > 0:
            memPercent = (self._dataMemApplications / 1024) / memTotal * 100
                
        return memPercent

    def getSwapPercent(self):
        swapTotal = (self._dataSwapUsed + self._dataSwapFree) / 1024
        swapUsage = self._dataSwapUsed / 1024
        swapPercent = 0.0
        if swapTotal > 0:
            swapPercent = swapUsage / swapTotal * 100
            
        return swapPercent

    def updateChart(self):
        memPercent = self.getMemPercent()
        if memPercent > 100:
            memPercent = 0

        swapPercent = self.getSwapPercent()
        if swapPercent > 100:
            swapPercent = 0
            
        # show cpu + memory usage as chart
        if self._showChart:
            samples = []
            
            if self._showCpu:
                samples.append(self._dataCpuUsed)
                
            if self._showMem:
                samples.append(memPercent)
                
            if self._showSwap:
                samples.append(swapPercent)
                
            # samples = [self._dataCpuUsed,memPercent,  swapPercent]
            self.chart.addSample(samples)


    def updateToolTip(self):
        # cpu
        cpuUsage = int( self._dataCpuUsed / 5) * 5
        cpuStr = 'CPU: ' + str(int(cpuUsage)) + "% @ " + str(self._dataCpuClock) + " Mhz per CPU<br/>"

        # mem
        memTotal = (self._dataMemFree + self._dataMemUsed) / 1024
        memUsed = self._dataMemApplications / 1024
        memPercent = self.getMemPercent()
        memStr = "Memory: " + str(int(memUsed))  + " MiB (" + str(int(memPercent)) + "%) of " \
               + str(int(memTotal)) + " MiB<br>" \
            
        #swap
        swapTotal = (self._dataSwapUsed + self._dataSwapFree) / 1024
        swapUsage = self._dataSwapUsed / 1024
        swapPercent = 0.0
        if swapTotal > 0:
            swapPercent = self.getSwapPercent()
        swapStr = "Swap: " + str(int(swapUsage)) + " MiB (" + str(int(swapPercent)) + "%) of " \
                + str(int(swapTotal)) + " Mib"

        # additional information
        extraStr = ""
        if self._showChart:
            extraStr = '<br><br> Plot: <font color="green">CPU</font>, <font color="red">Memory</font>, <font color="yellow">Swap</font>'
        
        # the string to show
        txt = cpuStr + memStr + swapStr + extraStr
        self.tooltipdata.setSubText(txt)
        
        Plasma.ToolTipManager.self().setContent(self.applet, self.tooltipdata)
            
    def toolTipAboutToShow(self):
        # TODO: this should be called before the tooltip is shown, but it isn't,
        # hence we call updateToolTip for each data update even if toolTip is not shown
        print("toolTipAboutToShow")
        self.updateToolTip()


    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.mousePressLocation = event.screenPos()                


    def mouseReleaseEvent(self, event):
        if event.button() != Qt.LeftButton:
            return

        # click or drag?
        if (event.screenPos() - self.mousePressLocation).manhattanLength() < QApplication.startDragDistance():
            # The following makes calls async, yet leads to crash for mutliple clicks
            # DBusQtMainLoop(set_as_default=True)  
            dbusSessionBus = dbus.SessionBus()
            dbusProxy = dbusSessionBus.get_object("org.kde.krunner", "/App")
            dbusProxy.showTaskManager(dbus_interface="org.kde.krunner.App")
                
    
        
def CreateApplet(parent):
    return CpuAndSystemViewer(parent)
